#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from overrides import overrides

from ..tweet.conversation_tweet_stream import ConversationTweetStream
from .conversation_request import ConversationRequest


class Replies(ConversationRequest):
    @overrides
    def request(self) -> ConversationTweetStream:
        from .._retriever.replies_retriever import RepliesRetriever

        return RepliesRetriever(self).tweet_stream
