_.error_on_external_run  # unused attribute (noxfile.py:20)
_.reuse_existing_virtualenvs  # unused attribute (noxfile.py:21)
_.stop_on_first_error  # unused attribute (noxfile.py:22)
test  # unused function (noxfile.py:25)
_max_tweets_validator  # unused function (src/nasty/_cli.py:74)
_batch_size_validator  # unused function (src/nasty/_cli.py:90)
title  # unused variable (src/nasty/_cli.py:129)
aliases  # unused variable (src/nasty/_cli.py:130)
description  # unused variable (src/nasty/_cli.py:131)
_since_validator  # unused function (src/nasty/_cli.py:150)
_until_validator  # unused function (src/nasty/_cli.py:161)
_daily_validator  # unused function (src/nasty/_cli.py:196)
title  # unused variable (src/nasty/_cli.py:236)
aliases  # unused variable (src/nasty/_cli.py:237)
description  # unused variable (src/nasty/_cli.py:238)
title  # unused variable (src/nasty/_cli.py:267)
aliases  # unused variable (src/nasty/_cli.py:268)
description  # unused variable (src/nasty/_cli.py:269)
title  # unused variable (src/nasty/_cli.py:297)
aliases  # unused variable (src/nasty/_cli.py:298)
description  # unused variable (src/nasty/_cli.py:299)
title  # unused variable (src/nasty/_cli.py:340)
aliases  # unused variable (src/nasty/_cli.py:341)
description  # unused variable (src/nasty/_cli.py:342)
_out_dir_validator  # unused function (src/nasty/_cli.py:367)
title  # unused variable (src/nasty/_cli.py:397)
aliases  # unused variable (src/nasty/_cli.py:398)
description  # unused variable (src/nasty/_cli.py:399)
_out_dir_validator  # unused function (src/nasty/_cli.py:426)
title  # unused variable (src/nasty/_cli.py:452)
version  # unused variable (src/nasty/_cli.py:453)
description  # unused variable (src/nasty/_cli.py:454)
subprograms  # unused variable (src/nasty/_cli.py:455)
_.num_tombstones  # unused attribute (src/nasty/_retriever/conversation_retriever.py:36)
_.num_tombstones  # unused attribute (src/nasty/_retriever/replies_retriever.py:139)
_.num_tombstones  # unused attribute (src/nasty/_retriever/thread_retriever.py:132)
SingleMetavarHelpFormatter  # unused class (src/nasty/_util/argparse_.py:23)
_._format_action_invocation  # unused method (src/nasty/_util/argparse_.py:24)
pytest_configure  # unused function (tests/conftest.py:30)
activate_requests_cache  # unused function (tests/conftest.py:56)
disrespect_robotstxt  # unused function (tests/conftest.py:67)
min_tombstones  # unused variable (tests/retriever/test_replies.py:70)
min_tombstones  # unused variable (tests/retriever/test_thread.py:67)
exc_type  # unused variable (tests/util/requests_cache.py:174)
exc_val  # unused variable (tests/util/requests_cache.py:175)
exc_tb  # unused variable (tests/util/requests_cache.py:176)
